;; Generated package description from diff-hl.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "diff-hl" "1.10.0.0.20260627.186" "Highlight uncommitted changes using VC" '((cl-lib "0.2") (emacs "27.1")) :commit "2d7d0714d9637f54af672987c65b6973b31e56a2" :authors '(("Dmitry Gutov" . "dmitry@gutov.dev")) :maintainer '("Dmitry Gutov" . "dmitry@gutov.dev") :keywords '("vc" "diff") :url "https://github.com/dgutov/diff-hl")
